"""
   
Updated and maintained by destination0b10unknown@gmail.com
Copyright 2022 destination2unknown

Licensed under the MIT License;
you may not use this file except in compliance with the License.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
   
"""
import threading
import time 
from scipy.integrate import odeint
import numpy as np

class sthread(threading.Thread):
   def __init__(self,func):
      threading.Thread.__init__(self)
      self.func=func
      
   def run(self):      
      self.func()

class PeriodicInterval(threading.Thread):
    def __init__(self, task_function, period):
        super().__init__()
        self.daemon = True
        self.task_function = task_function
        self.period = period
        self.i = 0
        self.t0 = time.time()
        self.stopper=0
        self.start()
        
    def sleep(self):
        self.i += 1
        delta = self.t0 + self.period * self.i - time.time()
        if delta > 0:
            time.sleep(delta)
    
    def run(self):
        while self.stopper==0:
            self.task_function()
            self.sleep()

    def stop(self):
        self.stopper=1

    def starter(self):
        self.stopper=0  
        self.i = 0
        self.t0 = time.time()

class tunefinderFOPDT(object):
    def __init__(self):
        self.Gain, self.TimeConstant, self.DeadTime = 1.5,60.0,15.0
        self.CHRKp,self.CHRKi, self.CHRKd=0.1,0.01,0.001
        self.IMCKp,self.IMCKi,self.IMCKd=0.2,0.02,0.002
        self.AIMCKp,self.AIMCKi, self.AIMCKd=0.3,0.03,0.003
            
    def calc(self,ModelData):
        self.Gain, self.TimeConstant, self.DeadTime = ModelData
        if (self.TimeConstant<=0):
            self.TimeConstant=1
        if (self.DeadTime<=0):
            self.DeadTime=1
        #CHRKp
        num=0.35*self.TimeConstant 
        den=abs(self.Gain)*self.DeadTime
        self.CHRKp=num/den
        #CHRKi
        ti=1.2*self.TimeConstant
        self.CHRKi=self.CHRKp/ti
        #CHRKd
        td=0.5*self.DeadTime
        self.CHRKd=(self.CHRKp*td)/60
                
        #IMC_Kp
        lmda=2.1*self.DeadTime
        num= self.TimeConstant+0.5*self.DeadTime
        den=abs(self.Gain)*(lmda)
        self.IMCKp = num/den
        #IMC_Ki
        ti=self.TimeConstant+0.5*self.DeadTime
        self.IMCKi = self.IMCKp / ti    
        #IMC_Kd
        num=self.TimeConstant*self.DeadTime
        den=2*self.TimeConstant+self.DeadTime
        td=num/den
        self.IMCKd = (td*self.IMCKp)/60
         
        #AIMC Kp
        L=max(0.1*self.TimeConstant,0.8*self.DeadTime)
        self.AIMCKp=self.TimeConstant/(abs(self.Gain)*(self.DeadTime+L))        
        #AIMC Ki
        ti=self.TimeConstant/(1.03-0.165*(self.DeadTime/self.TimeConstant))
        self.AIMCKi =self.AIMCKp/self.TimeConstant
        #AIMC Kd
        self.AIMCKd=(self.DeadTime/2)/60

    def calcFullFat(self,ModelData):
        self.Gain, self.TimeConstant, self.DeadTime = ModelData
        if (self.TimeConstant<=0):
            self.TimeConstant=1
        if (self.DeadTime<=0):
            self.DeadTime=1  
        #CHRKp
        num=0.6*self.TimeConstant 
        den=abs(self.Gain)*self.DeadTime
        self.CHRKp=num/den
        #CHRKi
        ti=1*self.TimeConstant
        self.CHRKi=self.CHRKp/ti
        #CHRKd
        td=0.5*self.DeadTime
        self.CHRKd=(self.CHRKp*td)
               
        #IMC_Kp
        lmda=2.1*self.DeadTime
        num= self.TimeConstant+0.5*self.DeadTime
        den=abs(self.Gain)*(lmda)
        self.IMCKp = 1.1*(num/den)
        #IMC_Ki
        ti=self.TimeConstant+0.5*self.DeadTime
        self.IMCKi = self.IMCKp / ti    
        #IMC_Kd
        num=self.TimeConstant*self.DeadTime
        den=2*self.TimeConstant+self.DeadTime
        td=num/den
        self.IMCKd = 1.1*(td*self.IMCKp)         

        #AIMC Kp
        L=max(0.1*self.TimeConstant,0.8*self.DeadTime)
        num=self.TimeConstant + 0.5*self.DeadTime
        den=abs(self.Gain)*(L+0.5*self.DeadTime)
        self.AIMCKp=num/den
        #AIMC Ki
        ti=self.TimeConstant+0.5*self.DeadTime
        self.AIMCKi=self.AIMCKp/ti
        #AIMC Kd
        num=self.TimeConstant*self.DeadTime
        den=2*self.TimeConstant+self.DeadTime
        td=num/den
        self.AIMCKd = td*self.AIMCKp

class PID(object):    
    def __init__(self,Kp=1.0,Ki=0.1,Kd=0.01,setpoint=50,output_limits=(0, 100)):
        self.Kp, self.Ki, self.Kd = Kp, Ki, Kd
        self.setpoint = setpoint
        self._min_output, self._max_output = 0, 100
        self._proportional = 0
        self._integral = 0
        self._derivative = 0
        self.output_limits = output_limits
        self._last_eD =0
        self._lastCV=0
        self._d_init=0      
        self.reset()

    def _clamp(self, value, limits):
        lower, upper = limits
        if value is None:
            return None
        elif (upper is not None) and (value > upper):
            return upper
        elif (lower is not None) and (value < lower):
            return lower
        return value

    def __call__(self,PV=0,SP=0,action="Direct",dFilter=1,pvt=1):
            #PID calculations            
            #P term
            if action=="Direct":
                e = SP - PV 
            else:
                e = PV - SP      
            self._proportional = self.Kp * e

            #I Term
            if self._lastCV<100 and self._lastCV >0:        
                self._integral += self.Ki * e
            #Allow I Term to change when Kp is set to Zero
            if self.Kp==0 and self._lastCV==100 and self.Ki * e<0:
                self._integral += self.Ki * e
            if self.Kp==0 and self._lastCV==0 and self.Ki * e>0:
                self._integral += self.Ki * e

            #D term
            if action=="Direct":
                eD=-PV 
            else:
                eD=PV            
            self._derivative = dFilter*self.Kd*(eD - self._last_eD)
            #init D term 
            if self._d_init==0:
                self._derivative=0
                self._d_init=1
            
            #pv tracking
            if pvt==0:
                self._integral=-(self.Kp*e)+0.00001
                
            #Controller Output
            CV = self._proportional + self._integral + self._derivative
            CV = self._clamp(CV, self.output_limits)

            # update stored data for next iteration
            self._last_eD = eD
            self._lastCV=CV
            return CV
        
    @property
    def components(self):
        return self._proportional, self._integral, self._derivative

    @property
    def tunings(self):
        return self.Kp, self.Ki, self.Kd

    @tunings.setter
    def tunings(self, tunings):        
        self.Kp, self.Ki, self.Kd = tunings
    
    @property
    def output_limits(self): 
        return self._min_output, self._max_output

    @output_limits.setter
    def output_limits(self, limits):        
        if limits is None:
            self._min_output, self._max_output = 0, 100
            return

        min_output, max_output = limits
        self._min_output = min_output
        self._max_output = max_output
        self._integral = self._clamp(self._integral, self.output_limits)    
    
    def reset(self):
        #Reset
        self._proportional = 0
        self._integral = 0
        self._derivative = 0
        self._integral = self._clamp(self._integral, self.output_limits)
        self._last_eD = 0
        self._lastCV = 0
        self._last_eD = 0
        self._d_init = 0

class FOPDTModel(object):   
    def __init__(self, CV, ModelData):                
        self.CV= CV
        self.Gain, self.TimeConstant, self.DeadTime, self.Bias = ModelData
        
    def calc(self,PV,ts):                      
        if (ts-self.DeadTime) <= 0:
            um=0
        elif int(ts-self.DeadTime)>=len(self.CV):
            um=self.CV[-1]
        else:
            um=self.CV[int(ts-self.DeadTime)]
        dydt = (-(PV-self.Bias) + self.Gain * um)/(self.TimeConstant)
        return dydt

    def update(self,PV,ts):
        y=odeint(self.calc,PV,ts)           
        return y[-1]


"""
   
Updated and maintained by destination0b10unknown@gmail.com
Copyright 2022 destination2unknown

Licensed under the MIT License;
you may not use this file except in compliance with the License.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
   
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import odeint
import tkinter as tk
from pytunelogix.common import generalclasses as g

#Random Noise between -0.25 and 0.25, same set used for each run. Created once at runtime.
minsize=600
maxsize=25200
noise= np.random.rand(maxsize)/2
noise-=0.25

def main():  
    def refresh():
        #get values from tkinter 
        if str(ind_button['state']) =='disabled' and str(sec_button['state'])=='disabled':
            calcKp.set(tKp.get())
            calcKi.set(tKi.get())
            calcKd.set(tKd.get())
        elif str(ind_button['state']) =='disabled' and str(sec_button['state'])=='normal': 
            calcKp.set(tKp.get())
            calcKi.set(float(tKi.get())/60)
            calcKd.set(float(tKd.get())*60)	
        elif str(ind_button['state']) =='normal' and str(sec_button['state'])=='disabled':
            calcKp.set(tKp.get())
            calcKi.set(float(tKp.get())/(float(tKi.get())))
            calcKd.set(float(tKp.get())*float(tKd.get()))
        elif str(ind_button['state']) =='normal' and str(sec_button['state'])=='normal':              
            calcKp.set(tKp.get())
            calcKi.set(float(tKp.get())/(float(tKi.get())*60))
            calcKd.set(float(tKp.get())*float(tKd.get())*60)
        igain,itau,ideadtime=float(tK.get()),float(ttau.get()),float(tdt.get())
        ikp,iki,ikd = float(calcKp.get()),float(calcKi.get()),float(calcKd.get())
            
        #Find the size of the range needed
        if (ideadtime+itau)*7 < minsize:
         rangesize = minsize
        elif (ideadtime+itau)*7 >maxsize:
         rangesize = maxsize
        else:
         rangesize = int((ideadtime+itau)*7)

        #setup time intervals
        t = np.arange(start=0, stop=rangesize, step=1)

        #Setup data arrays
        SP = np.zeros(len(t)) 
        PV = np.zeros(len(t))
        CV = np.zeros(len(t))
        pterm = np.zeros(len(t))
        iterm = np.zeros(len(t))
        dterm = np.zeros(len(t))
        global noise
        noise=np.resize(noise, len(t))
        
        #defaults
        ibias=float(tAmb.get())
        startofstep=10

        #Packup data
        PIDGains=(ikp,iki,ikd)
        ModelData=(igain,itau,ideadtime,ibias)
        #Filter for D term
        if radioFilter.get()=='Filter':
            tc=max(0.1*itau,0.8,ideadtime)
            num=tc*(itau+0.5*ideadtime)
            den=itau*(tc+ideadtime)
            dFilter=num/den
        else:
            dFilter=1
        direction=radioDirection.get()
        
        #Process Variable Tracking
        if radioPvT.get()=='PvT':
            PvT=1
        else:
            PvT=0

        #PID Instantiation
        pid = g.PID(ikp, iki, ikd, SP[0])
        pid.output_limits = (0, 100)
        pid.tunings=(PIDGains)

        #plant Instantiation
        plant=g.FOPDTModel(CV, ModelData)

        #Start Value
        PV[0]=ibias+noise[0]
        
        #Loop through to find PID output and Process value
        for i in t:        
            if i<len(t)-1:            
                if i < startofstep-1:
                    SP[i] = ibias
                elif i< rangesize*0.6:
                    if direction=="Direct":
                        SP[i]= 60 + ibias
                    else:
                        SP[i] = ibias - 60
                else:
                    if direction=="Direct":
                        SP[i]= 40 + ibias
                    else:
                        SP[i]= ibias - 40
                #Find current controller output                
                CV[i]=pid(PV[i], SP[i],direction,dFilter,0 if (i < startofstep and PvT==0) else 1)
                ts = [t[i],t[i+1]]
                #Send step data
                plant.CV=CV
                #Find calculated PV
                PV[i+1] = plant.update(PV[i],ts)
                PV[i+1]+=noise[i]
                #Store indiv. terms
                pterm[i],iterm[i],dterm[i]=pid.components
            else:
                #cleanup endpoint
                SP[i]=SP[i-1]
                CV[i]=CV[i-1]
                pterm[i]=pterm[i-1]
                iterm[i]=iterm[i-1]
                dterm[i]=dterm[i-1]
            itae = 0 if i < startofstep else itae+(i-startofstep)*abs(SP[i]-PV[i])
                
        #Display itae value    
        itae_text.set(round(itae/len(t),2)) #measure PID performance
        
        #Plots
        plt.figure()    
        plt.subplot(2, 1, 1) 
        plt.plot(t,SP, color="blue", linewidth=2, label='SP')
        plt.plot(t,CV,color="darkgreen",linewidth=2,label='CV')
        plt.plot(t,PV,color="red",linewidth=2,label='PV')    
        plt.ylabel('EU')    
        plt.suptitle("ITAE: %s" % round(itae/len(t),2))        
        plt.title(tUnitP.get()+": "+tKp.get() + "      " + tUnitI.get()+ ": " + tKi.get()+"      " + tUnitD.get() +": "   + tKd.get() ,fontsize=10)
        plt.legend(loc='best')

        plt.subplot(2,1,2)
        plt.plot(t,pterm, color="lime", linewidth=2, label='P Term')
        plt.plot(t,iterm,color="orange",linewidth=2,label='I Term')
        plt.plot(t,dterm,color="purple",linewidth=2,label='D Term')        
        plt.xlabel('Time [seconds]')
        plt.legend(loc='best')
        plt.show(block=False)

    def convert(selector):   
        if selector == "S":
            sec_button['state']='disabled'
            min_button['state']='normal'
            if str(dep_button['state']) =='normal' and str(ind_button['state'])=='disabled':
                tUnitP.set("Kp")
                tUnitI.set("Ki (1/s)")
                tUnitD.set("Kd (s)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKi.get())/60,6))
                thisd=str(round(float(tKd.get())*60,6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

            elif str(dep_button['state']) =='disabled' and str(ind_button['state'])=='normal':
                tUnitP.set("Kc")
                tUnitI.set("Ti (sec/repeat)")
                tUnitD.set("Td (sec)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKi.get())*60,6))
                thisd=str(round(float(tKd.get())*60,6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

        elif selector == "M":
            sec_button['state']='normal'
            min_button['state']='disabled'
            if str(dep_button['state']) =='normal' and str(ind_button['state'])=='disabled':
                tUnitP.set("Kp")
                tUnitI.set("Ki (1/min)")
                tUnitD.set("Kd (min)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKi.get())*60,6))
                thisd=str(round(float(tKd.get())/60,6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

            elif str(dep_button['state']) =='disabled' and str(ind_button['state'])=='normal':
                tUnitP.set("Kc")
                tUnitI.set("Ti (min/repeat)")
                tUnitD.set("Td (min)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKi.get())/60,6))
                thisd=str(round(float(tKd.get())/60,6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

        elif selector == "D":
            ind_button['state']='normal'
            dep_button['state']='disabled'
            if str(sec_button['state']) =='disabled' and str(min_button['state'])=='normal':              
                tUnitP.set("Kc")
                tUnitI.set("Ti (sec/repeat)")
                tUnitD.set("Td (sec)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKp.get())/float(tKi.get()),6))
                thisd=str(round(float(tKd.get())/float(tKp.get()),6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

            elif str(sec_button['state']) =='normal' and str(min_button['state'])=='disabled':              
                tUnitP.set("Kc")
                tUnitI.set("Ti (min/repeat)")
                tUnitD.set("Td (min)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKp.get())/(float(tKi.get())),6))
                thisd=str(round(float(tKd.get())/float(tKp.get()),6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

        elif selector == "I":
            ind_button['state']='disabled'
            dep_button['state']='normal'
            if str(sec_button['state']) =='disabled' and str(min_button['state'])=='normal':                    
                tUnitP.set("Kp")
                tUnitI.set("Ki (1/s)")
                tUnitD.set("Kd (s)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKp.get())/float(tKi.get()),6))
                thisd=str(round(float(tKd.get())*float(tKp.get()),6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)

            elif str(sec_button['state']) =='normal' and str(min_button['state'])=='disabled': 
                tUnitP.set("Kp")
                tUnitI.set("Ki (1/min)")
                tUnitD.set("Kd (min)")
                thisp=str(round(float(tKp.get()),6))
                thisi=str(round(float(tKp.get())/float(tKi.get()),6))
                thisd=str(round(float(tKd.get())*float(tKp.get()),6))
                tKp.delete(0,'end')
                tKi.delete(0,'end')
                tKd.delete(0,'end')
                tKp.insert(0,thisp)
                tKi.insert(0,thisi)
                tKd.insert(0,thisd)
        
    #Gui
    root = tk.Tk()
    root.title('PID Simulator with a FOPDT Process Model')
    root.resizable(True, True)
    root.geometry('500x180')

    #Labels
    tK = tk.Entry(root,width=8)
    ttau = tk.Entry(root,width=8)
    tdt= tk.Entry(root,width=8)
    tKp = tk.Entry(root,width=8)
    tKi = tk.Entry(root,width=8)
    tKd= tk.Entry(root,width=8)
    tAmb = tk.Entry(root,width=8)
    radioFilter = tk.StringVar()
    radioDirection = tk.StringVar()
    radioPvT = tk.StringVar()
    itae_text = tk.StringVar()
    tUnitP = tk.StringVar()
    tUnitI = tk.StringVar()
    tUnitD = tk.StringVar()
    tUnits = tk.StringVar()
    calcKp = tk.StringVar()
    calcKi = tk.StringVar()
    calcKd = tk.StringVar()

    tk.Label(root, text=" ").grid(row=0,column=0)
    tk.Label(root, text="FOPDT").grid(row=0,column=1)
    tk.Label(root, text="Model Gain: ").grid(row=1,sticky="E")
    tk.Label(root, text="TimeConstant: ").grid(row=2,sticky="E")
    tk.Label(root, text="DeadTime: ").grid(row=3,sticky="E")
    tk.Label(root, text="Ambient: ").grid(row=4,sticky="E")
    tk.Label(root, text="                 ").grid(row=0,column=2,sticky="W")
    tk.Label(root, text="                 ").grid(row=0,column=3,sticky="W")
    tk.Label(root, text="PID Gains").grid(row=0,column=4)
    tk.Label(root, textvariable=tUnitP).grid(row=1,column=2,columnspan=2,sticky="E")
    tk.Label(root, textvariable=tUnitI).grid(row=2,column=2,columnspan=2,sticky="E")
    tk.Label(root, textvariable=tUnitD).grid(row=3,column=2,columnspan=2,sticky="E")

    tK.insert(0, "2.25")
    ttau.insert(0, "60.5")
    tdt.insert(0, "9.99")
    tKp.insert(0, "3")
    tKi.insert(0, "0.1")
    tKd.insert(0, "3")
    calcKp.set("3")
    calcKi.set("0.1")
    calcKd.set("3")
    tAmb.insert(0, "13.5")
    radioFilter.set("Filter")
    radioDirection.set("Direct")
    radioPvT.set("PvT")
    itae_text.set("...")
    tUnitP.set("Kp")
    tUnitI.set("Ki (1/s)")
    tUnitD.set("Kd (s)")

    tK.grid(row=1, column=1)
    ttau.grid(row=2, column=1)
    tdt.grid(row=3, column=1)
    tKp.grid(row=1, column=4)
    tKi.grid(row=2, column=4)
    tKd.grid(row=3, column=4)
    tAmb.grid(row=4, column=1)

    button_refresh = tk.Button(root, text="Refresh", command=refresh)
    button_refresh.grid(row=5,column=1)
    tk.Label(root, text="itae:").grid(row=4,column=3)
    tk.Label(root, textvariable=itae_text).grid(row=4,column=4,padx=5,sticky="NESW")
    
    tk.Radiobutton(root, text = "Filter", variable=radioFilter, value = "Filter").grid(row=3,column=7,padx=5,pady=0, sticky="NESW")
    tk.Radiobutton(root, text = "UnFiltered", variable=radioFilter, value = "UnFilter").grid(row=3,column=8,padx=5,pady=0,columnspan=2, sticky="NESW")
    tk.Radiobutton(root, text = "Direct", variable=radioDirection, value = "Direct").grid(row=4,column=7,padx=5,pady=0, sticky="NESW")
    tk.Radiobutton(root, text = "Reverse", variable=radioDirection, value = "Reverse").grid(row=4,column=8,padx=5,pady=0,columnspan=2, sticky="NESW")
    tk.Radiobutton(root, text = "PV Tracking", variable=radioPvT, value = "PvT").grid(row=5,column=7,padx=5,pady=0, sticky="NESW")
    tk.Radiobutton(root, text = "No PvT", variable=radioPvT, value = "NoPvT").grid(row=5,column=8,padx=5,pady=0,columnspan=2, sticky="NESW")

    ind_button = tk.Button(root,text='Independant',command=lambda :[convert("I")])
    ind_button.grid(row=1,column=7,padx=5,pady=0,sticky="NESW")
    ind_button['state']='disabled'

    dep_button = tk.Button(root,text='Dependant',command=lambda :[convert("D")])
    dep_button.grid(row=1,column=8,padx=5,pady=0,sticky="NESW")

    sec_button = tk.Button(root,text='Seconds',command=lambda :[convert("S")])
    sec_button.grid(row=2,column=7,padx=5,pady=0,sticky="NESW")
    sec_button['state']='disabled'

    min_button = tk.Button(root,text='Minutes',command=lambda :[convert("M")])
    min_button.grid(row=2,column=8,padx=5,pady=0,sticky="NESW")

    root.mainloop()
 
if __name__ == '__main__':
    main()


"""
   
Updated and maintained by destination0b10unknown@gmail.com
Copyright 2022 destination2unknown

Licensed under the MIT License;
you may not use this file except in compliance with the License.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
   
"""
import sys
import os
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from scipy import signal
import numpy as np
import tkinter as tk
from tkinter import ttk
from tkinter.messagebox import showinfo
import tkinter.filedialog
from pytunelogix.common import generalclasses as g

def main():
    # create the root window
    root = tk.Tk()
    root.title('PID Tuner Based on a FOPDT Model')
    root.resizable(True, True)
    root.geometry('550x380')

    fname = tk.StringVar()
    modelgain = tk.StringVar()
    modeltc = tk.StringVar()
    modeldt = tk.StringVar()
    ambient = tk.StringVar()
    tCHRKp = tk.StringVar()
    tCHRKi = tk.StringVar()
    tCHRKd = tk.StringVar()
    tIMCKp = tk.StringVar()
    tIMCKi = tk.StringVar()
    tIMCKd = tk.StringVar()
    tAIMCKp = tk.StringVar()
    tAIMCKi = tk.StringVar()
    tAIMCKd = tk.StringVar()
    tUnitKp = tk.StringVar()
    tUnitKi= tk.StringVar()
    tUnitKd= tk.StringVar()

    def step1():
        filetypes = (('CSV files', '*.csv'),('All files', '*.*'))
        filename= tk.filedialog.askopenfilename(
            title='Open a file',
            initialdir=os.getcwd(),
            filetypes=filetypes)
        if filename!="":
            fname.set(filename)
                
    def step2(filename):
        if bool(fname.get()) == True:
            #Calculation Variables
            deltaT=0.1 #assume 100ms interval
            twosecwindow=int(2/deltaT)
            onesecwindow=int(1/deltaT)
            hihilimit=1.05
            lololimit=0.95
            hilimit=1.02
            lowlimit=0.98
            settlingwindow=5
            window=int(settlingwindow/deltaT)

            df = pd.read_csv(filename, sep=';')
            #Find relevant columns
            CV_cols = [col for col in df.columns if 'CV' in col.upper()]
            PV_cols = [col for col in df.columns if 'PV' in col.upper()]

            #Creat DataFrames
            df['CV'] = df[CV_cols[0]]
            df['PV'] = df[PV_cols[0]]

            #Direction
            if df['PV'].iloc[1] > df['PV'].iloc[-1]:
                direction="Reverse"
            else:
                direction="Direct"

            #Find basic parameters
            i_start=df['CV'].idxmax()
            ambient.set(round(df['PV'].iloc[i_start:i_start+(twosecwindow*4):1].mean(axis = 0),2))
            StartPV=df['PV'].iloc[i_start]
            InitCV=df['CV'].iloc[0]
            StartCV=df['CV'].iloc[i_start]

            #DeadTime Range
            RangeU=round(float(ambient.get()),2)*hihilimit
            RangeL=round(float(ambient.get()),2)*lololimit
                    
            #Find DeadTime
            if direction=="Reverse":
                for x in range(i_start,(len(df['PV'])-window)):
                    if(df['PV'].iloc[x:x+twosecwindow:1].mean(axis = 0)<RangeL):
                        modeldt.set(round((x-i_start)*deltaT,2))
                        break
            else:
                for x in range(i_start,(len(df['PV'])-window)):
                    if(df['PV'].iloc[x:x+twosecwindow:1].mean(axis = 0)>RangeU):
                        modeldt.set(round((x-i_start)*deltaT,2))
                        break

            #Find Gain
            if direction=="Reverse":
                j=df['PV'].idxmin()
                min_peak=df['PV'].min()
            else:
                j=df['PV'].idxmax()
                max_peak=df['PV'].max()
                
            peak = df['PV'].iloc[j-onesecwindow:j+onesecwindow:1].mean(axis = 0)
            modelgain.set(round((peak-round(float(ambient.get()),2))/(StartCV-InitCV),2))

            #Time Constant
            tc_value=0.63*(peak-float(ambient.get()))+float(ambient.get())
            tc_upp=tc_value*1.01
            tc_low=tc_value*0.99

            #Find Time Constant
            if direction=="Reverse": 
                z=df[df['PV']<=StartPV-(peak*0.1)].first_valid_index()        
                for x in range(z,(len(df['PV'])-window)):
                    if(df['PV'].iloc[x-twosecwindow:x+twosecwindow:1].mean(axis = 0)<tc_low):
                        modeltc.set(round(((x-i_start)*deltaT)-float(modeldt.get()),2))
                        break
                    else:
                        modeltc.set(-1)
                        
            else:
                z=df[df['PV']>=(peak*0.5)].first_valid_index()        
                for x in range(z,(len(df['PV'])-window)):
                    if(df['PV'].iloc[x-twosecwindow:x+twosecwindow:1].mean(axis = 0)>tc_value):
                        modeltc.set(round(((x-i_start)*deltaT)-float(modeldt.get()),2))
                        break
                    else:
                        modeltc.set(-1)

            #Setup System Model
            num = [float(modelgain.get())]
            den = [float(modeltc.get()),1]
            sys1 = signal.TransferFunction(num,den)
            tune() 

            #Find Step Response based on Rough Model
            t1,y1 = signal.step(sys1, N=int(df['PV'].count()/10))
            t1=t1[::10]
            y1=y1[::10]

            #Rescale
            plotpv=df['PV'].iloc[::10].reset_index(drop=True)
            plotcv=df['CV'].iloc[::10].reset_index(drop=True)

            #plot 
            plt.figure()
            plt.xlim(0,df['PV'].count()*0.11)
            plt.plot(plotpv, color="blue", linewidth=3, label='PV')
            plt.plot(t1+(i_start/10+(float(modeldt.get()))),((StartCV-InitCV)*y1)+round(float(ambient.get()),2),color="red",linewidth=3,label='Model')
            plt.plot(plotcv, color="green", linewidth=3, label='CV')
            plt.hlines(round(float(ambient.get()),2), 0, i_start/10+float(modeldt.get()),colors='red', linestyles='solid',linewidth=3,label='')
            plt.ylabel('Engineering Units')
            plt.xlabel('Seconds')
            plt.suptitle("Rough Model v Actual Data")
            plt.title(f"ModelGain: {round(float(modelgain.get()),2)}   ModelTc: {round(float(modeltc.get()),2)}   ModelDT: {round(float(modeldt.get()),2)}")
            plt.legend(loc='best')
            plt.show(block=False)

    def step3(filename):
        if bool(fname.get())==True:
            #Read File 
            df = pd.read_csv(filename, sep=';')

            #Find CV and PV Columns
            CV_cols = [col for col in df.columns if 'CV' in col.upper()]
            PV_cols = [col for col in df.columns if 'PV' in col.upper()]
            df['CV'] = df[CV_cols[0]]
            df['PV'] = df[PV_cols[0]]

            #Find Step Size
            CVStep=df['CV'].max()-df['CV'].min()
            indexofstart=int(df['CV'].idxmax()/10) #start of step
            ambient.set(round(df['PV'].iloc[int(indexofstart/2):indexofstart:1].mean(axis = 0),2))

            #produces a model based on the parameters
            def fopdt_func(t_fopdt, K=1, tau=1, deadtime=0):
                deadtime = max(0,deadtime)
                tau = max(0,tau)
                return np.array([K*(1-np.exp(-(t_fopdt-deadtime)/tau)) if t_fopdt >= deadtime else 0 for t_fopdt in t_fopdt])

            #Difference between model and actual
            def err(Xe,te,ye):
                Ke,tau,DeadTime = Xe
                z = CVStep*fopdt_func(te,Ke,tau,DeadTime)+float(ambient.get())
                iae = sum(abs(z-ye))*(max(te)-min(te))/len(te)
                return iae

            #Trim Timescale
            actualPV=df['PV'].iloc[(indexofstart*10)-1::10].reset_index(drop=True)
            actualCV=df['CV'].iloc[(indexofstart*10)-1::10].reset_index(drop=True)
            t=actualPV.index.values

            #Model Starting Point
            if modelgain.get() =="...":
                ModelValues = 1.1,60,10
            else:
                ModelValues = float(modelgain.get()),float(modeltc.get()), float(modeldt.get())
                
            #minimize difference between model and actual
            bounds = [(None, None), (0, None), (0,None)]
            Gain,Tau,DeadTime = minimize(err,ModelValues,args=(t, actualPV.values),bounds=bounds, method='Nelder-Mead').x

            #Update holder
            modelgain.set(round(Gain,2))
            modeltc.set(round(Tau,2))
            modeldt.set(round(DeadTime,2))
            tune()

            #Get data to plot new model 
            ymodel=CVStep*fopdt_func(t,float(modelgain.get()),float(modeltc.get()), float(modeldt.get()))+float(ambient.get())

            #Plot
            plt.figure()
            plt.plot(actualPV,color="blue",linewidth=3,label='PV')
            plt.plot(actualCV,color="green",linewidth=3,label='CV')
            plt.plot(ymodel,color="red",linewidth=3,label='Model')
            plt.xlabel('Seconds')
            plt.ylabel('Engineering Units')
            plt.suptitle('Refined Model v Actual Data')
            plt.title(f"ModelGain: {round(float(modelgain.get()),2)}   ModelTc: {round(float(modeltc.get()),2)}   ModelDT: {round(float(modeldt.get()),2)}")
            plt.legend(loc='best')
            plt.show(block=False)
                
    def runthesim(processmodel,tune):
        #EntryPoint
        minsize=2400
        maxsize=7200

        #Find the size of the range needed
        if float(modeldt.get())+float(modeltc.get())*6 < minsize:
            rangesize = minsize
        elif float(modeldt.get())+float(modeltc.get())*6 >maxsize:
            rangesize = maxsize
        else:
            rangesize = int((ideadtime+itau)*6)

        #setup time intervals
        t = np.arange(rangesize)
        #Random Noise between -0.1 and 0.1
        noise= 0.2*np.random.rand(rangesize)
        noise-=0.1
        #noise=np.zeros(rangesize) #no noise

        #unpack values 
        igain,itau,ideadtime=processmodel
        ikp,iki,ikd = tune

        #Setup data arrays
        SP = np.zeros(len(t)) 
        PV = np.zeros(len(t))
        CV = np.zeros(len(t))
        pterm = np.zeros(len(t))
        iterm = np.zeros(len(t))
        dterm = np.zeros(len(t))
        
        #defaults
        startofstep=10

        #Packup data
        PIDGains=(ikp,iki,ikd)
        ModelData=(igain,itau,ideadtime,float(ambient.get()))

        #PID Instantiation
        pid = g.PID(ikp, iki, ikd, SP[0])
        pid.output_limits = (0, 100)
        pid.tunings=(PIDGains)

        #plant Instantiation
        plant=g.FOPDTModel(CV, ModelData)

        #Start Value
        PV[0]=float(ambient.get())+noise[0]
        
        #Loop through timestamps
        for i in t:        
            if i<(len(t)-1):
                if i < startofstep:
                    SP[i] = float(ambient.get())
                elif (i > startofstep and i< rangesize/2):
                    if float(modelgain.get()) > 0:
                        SP[i]= 50 + float(ambient.get())
                    else:
                        SP[i]= float(ambient.get()) - 50
                else:
                    if float(modelgain.get()) > 0:
                        SP[i]=40 + float(ambient.get())
                    else:
                        SP[i]= float(ambient.get()) - 40
                #Find current controller output
                CV[i]=pid(PV[i], SP[i], "Direct" if float(modelgain.get()) > 0 else "Reverse")               
                ts = [i,i+1]
                #Send step data
                plant.CV=CV
                #Find calculated PV
                PV[i+1] = plant.update(PV[i],ts)
                PV[i+1]+=noise[i]
            else:
                #cleanup endpoint
                SP[i]=SP[i-1]
                CV[i]=CV[i-1]
            itae = 0 if i < startofstep else itae+(i-startofstep)*abs(SP[i]-PV[i])
                
        #Display itae value    
        itae=(round(itae/len(t),2)) #measure PID performance
        dataout=SP,PV,CV,itae
        return dataout

    def step4():
        if modelgain.get()!="..." and tCHRKp.get()!="...":
            processmodel=float(modelgain.get()),float(modeltc.get()), float(modeldt.get())
            CHRtune=tuner.CHRKp,tuner.CHRKi,tuner.CHRKd
            IMCtune= tuner.IMCKp,tuner.IMCKi,tuner.IMCKd
            AIMCtune=tuner.AIMCKp,tuner.AIMCKi,tuner.AIMCKd

            #store pid data array from simulation
            CHR=runthesim(processmodel,CHRtune)
            IMC=runthesim(processmodel,IMCtune)
            AIMC=runthesim(processmodel,AIMCtune)

            #Unpack the data returned
            CHRSP,CHRPV,CHRCV,CHRitae=CHR
            IMCSP,IMCPV,IMAIMCV,IMCitae=IMC
            AIMCSP,AIMCPV,AIMCCV,AIMCitae=AIMC

            plt.figure()    
            plt.plot(CHRSP, color="goldenrod", linewidth=3, label='SP')
            plt.plot(CHRPV,color="darkgreen",linewidth=2,label='CHR PV')
            plt.plot(IMCPV,color="blue",linewidth=2,label='IMC PV')    
            plt.plot(AIMCPV, color="red", linewidth=2, label='AIMC PV')
        
            plt.ylabel('EU')    
            plt.xlabel('Seconds')
            plt.suptitle("PID Tune Comparison")        
            plt.title("CHR ITAE:%s      IMC ITAE:%s      AIMC ITAE:%s" % (CHRitae, IMCitae, AIMCitae),fontsize=10)
            plt.legend(loc='best')
            plt.show(block=False)

    def tune(selector="C"):
        if modelgain.get()!="...":
            model=float(modelgain.get()),float(modeltc.get()), float(modeldt.get())
            if selector=="C":            
                con_button['state']='disabled'
                agg_button['state']='normal'
                tuner.calc(model)
            elif selector=='A':            
                con_button['state']='normal'
                agg_button['state']='disabled'
                tuner.calcFullFat(model)       
            if selector == "S":
                sec_button['state']='disabled'
                min_button['state']='normal'
            elif selector == "M":
                sec_button['state']='normal'
                min_button['state']='disabled'
            elif selector == "D":
                ind_button['state']='normal'
                dep_button['state']='disabled'
            elif selector == "I":
                ind_button['state']='disabled'
                dep_button['state']='normal'

            if str(dep_button['state']) =='normal' and str(sec_button['state'])=='disabled':
                tUnitKp.set("Kp")
                tUnitKi.set("      Ki (1/s)      ")
                tUnitKd.set("    Kd (s)    ")
                tCHRKp.set(round(tuner.CHRKp,4))
                tCHRKi.set(round(tuner.CHRKi,4))
                tCHRKd.set(round(tuner.CHRKd,4))
                tIMCKp.set(round(tuner.IMCKp,4))
                tIMCKi.set(round(tuner.IMCKi,4))
                tIMCKd.set(round(tuner.IMCKd,4))
                tAIMCKp.set(round(tuner.AIMCKp,4))
                tAIMCKi.set(round(tuner.AIMCKi,4))
                tAIMCKd.set(round(tuner.AIMCKd,4))

            elif str(dep_button['state']) =='normal' and str(sec_button['state'])=='normal':
                tUnitKp.set("Kp")
                tUnitKi.set("    Ki (1/min)    ")
                tUnitKd.set("   Kd (min)   ")
                tCHRKp.set(round(tuner.CHRKp,4))
                tCHRKi.set(round(tuner.CHRKi*60,4))
                tCHRKd.set(round(tuner.CHRKd/60,4))
                tIMCKp.set(round(tuner.IMCKp,4))
                tIMCKi.set(round(tuner.IMCKi*60,4))
                tIMCKd.set(round(tuner.IMCKd/60,4))
                tAIMCKp.set(round(tuner.AIMCKp,4))
                tAIMCKi.set(round(tuner.AIMCKi*60,4))
                tAIMCKd.set(round(tuner.AIMCKd/60,4))
            
            elif str(dep_button['state']) =='disabled' and str(sec_button['state'])=='disabled':
                tUnitKp.set("Kc")
                tUnitKi.set(" Ti (sec/repeat) ")
                tUnitKd.set("Td (sec)")
                tCHRKp.set(round(tuner.CHRKp,4))
                tCHRKi.set(round(tuner.CHRKp/(tuner.CHRKi),4))
                tCHRKd.set(round(tuner.CHRKd/(tuner.CHRKp),4))
                tIMCKp.set(round(tuner.IMCKp,4))
                tIMCKi.set(round(tuner.IMCKp/(tuner.IMCKi),4))
                tIMCKd.set(round(tuner.IMCKd/(tuner.IMCKp),4))
                tAIMCKp.set(round(tuner.AIMCKp,4))
                tAIMCKi.set(round(tuner.AIMCKp/(tuner.AIMCKi),4))
                tAIMCKd.set(round(tuner.AIMCKd/(tuner.AIMCKp),4))

            else: 
                tUnitKp.set("Kc")
                tUnitKi.set("Ti (min/repeat)")
                tUnitKd.set("Td (min)")
                tCHRKp.set(round(tuner.CHRKp,4))
                tCHRKi.set(round(tuner.CHRKp/(tuner.CHRKi*60),4))
                tCHRKd.set(round(tuner.CHRKd/(tuner.CHRKp*60),4))
                tIMCKp.set(round(tuner.IMCKp,4))
                tIMCKi.set(round(tuner.IMCKp/(tuner.IMCKi*60),4))
                tIMCKd.set(round(tuner.IMCKd/(tuner.IMCKp*60),4))
                tAIMCKp.set(round(tuner.AIMCKp,4))
                tAIMCKi.set(round(tuner.AIMCKp/(tuner.AIMCKi*60),4))
                tAIMCKd.set(round(tuner.AIMCKd/(tuner.AIMCKp*60),4))

    tuner=g.tunefinderFOPDT()
    modeldt.set('...')
    modeltc.set('...')
    modelgain.set('...')
    ambient.set('...')
    tCHRKp.set("...")
    tCHRKi.set("...")
    tCHRKd.set("...")
    tIMCKp.set("...")
    tIMCKi.set("...")
    tIMCKd.set("...")
    tAIMCKp.set("...")
    tAIMCKi.set("...")
    tAIMCKd.set("...")
    tUnitKp.set("     Kp     ")
    tUnitKi.set("      Ki (1/s)      ")
    tUnitKd.set("   Kd (s)   ")

    step1_button = ttk.Button(root,text='Step 1: Open CSV File    ',command=lambda :[step1()])
    step1_button.grid(row=0,column=0,padx=5,pady=4,sticky="NESW")

    step2_button = ttk.Button(root,text='Step 2: Estimate Model ',command=lambda :[step2(fname.get())])
    step2_button.grid(row=1,column=0,padx=5,pady=4,sticky="NESW")

    step3_button = ttk.Button(root,text='Step 3: Refine Model     ', command=lambda :[step3(fname.get())])
    step3_button.grid(row=2,column=0,padx=5,pady=4,sticky="NESW")

    step4_button = ttk.Button(root,text='Step 4: Run PID Sim       ',command=lambda :[step4()])
    step4_button.grid(row=3,column=0,padx=5,pady=4,sticky="NESW")

    con_button = ttk.Button(root,text='Conservative',command=lambda :[tune("C")])
    con_button.grid(row=6,column=0,padx=5,pady=4,sticky="NESW")
    con_button['state']='disabled'

    agg_button = ttk.Button(root,text='Aggressive',command=lambda :[tune("A")])
    agg_button.grid(row=7,column=0,padx=5,pady=4,sticky="NESW")

    ind_button = ttk.Button(root,text='Independant',command=lambda :[tune("I")])
    ind_button.grid(row=8,column=0,padx=5,pady=4,sticky="NESW")
    ind_button['state']='disabled'

    dep_button = ttk.Button(root,text='Dependant',command=lambda :[tune("D")])
    dep_button.grid(row=9,column=0,padx=5,pady=4,sticky="NESW")

    sec_button = ttk.Button(root,text='Seconds',command=lambda :[tune("S")])
    sec_button.grid(row=10,column=0,padx=5,pady=4,sticky="NESW")
    sec_button['state']='disabled'

    min_button = ttk.Button(root,text='Minutes',command=lambda :[tune("M")])
    min_button.grid(row=11,column=0,padx=5,pady=4,sticky="NESW")

    tk.Label(root, textvariable=fname).grid(row=0,column=1,columnspan=5,padx=1,pady=1,sticky="W")
    tk.Label(root, text="Model Gain:").grid(row=1,column=1,columnspan=2,padx=1,pady=1,sticky="W")
    tk.Label(root, text="TimeConstant (s):").grid(row=2,column=1,columnspan=1,padx=1,pady=1,sticky="W")
    tk.Label(root, text="DeadTime (s):").grid(row=3,column=1,columnspan=1,padx=1,pady=1,sticky="W")
    tk.Label(root, text="Ambient:").grid(row=4,column=1,columnspan=1,padx=1,pady=1,sticky="W")
    tk.Label(root, textvariable=modelgain).grid(row=1,column=2,padx=1,pady=1,sticky="NESW")
    tk.Label(root, textvariable=modeltc).grid(row=2,column=2,padx=1,pady=1,sticky="NESW")
    tk.Label(root, textvariable=modeldt).grid(row=3,column=2,padx=1,pady=1,sticky="NESW")
    tk.Label(root, textvariable=ambient).grid(row=4,column=2,padx=1,pady=1,sticky="NESW")
    spacer="____________________________________________________"
    tk.Label(root, text=spacer).grid(row=5,column=1,columnspan=4)
    tk.Label(root, text="PID Gains ->").grid(row=6,column=1,sticky="W")
    tk.Label(root, textvariable=tUnitKp).grid(row=6,column=2,sticky="NESW")
    tk.Label(root, textvariable=tUnitKi).grid(row=6,column=3,sticky="NESW")
    tk.Label(root, textvariable=tUnitKd).grid(row=6,column=4,sticky="NESW")
    tk.Label(root, text="CHR Method").grid(row=7,column=1,sticky="W")
    tk.Label(root, text="IMC Method").grid(row=8,column=1,sticky="W")
    tk.Label(root, text="AIMC Method").grid(row=9,column=1,sticky="W")
    tk.Label(root, textvariable=tCHRKp).grid(row=7,column=2,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tCHRKi).grid(row=7,column=3,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tCHRKd).grid(row=7,column=4,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tIMCKp).grid(row=8,column=2,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tIMCKi).grid(row=8,column=3,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tIMCKd).grid(row=8,column=4,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tAIMCKp).grid(row=9,column=2,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tAIMCKi).grid(row=9,column=3,padx=4,pady=4,sticky="NESW")
    tk.Label(root, textvariable=tAIMCKd).grid(row=9,column=4,padx=4,pady=4,sticky="NESW")

    # run the gui
    root.mainloop()

if __name__ == '__main__':
    main()


"""
   
Updated and maintained by destination0b10unknown@gmail.com
Copyright 2022 destination2unknown

Licensed under the MIT License;
you may not use this file except in compliance with the License.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
   
"""
import tkinter as tk
import numpy as np
import matplotlib.pyplot as plt
import random
import threading
import time
from matplotlib import animation
from scipy.integrate import odeint
from pylogix import PLC
from pytunelogix.common import generalclasses as g

def main():
    class data(object):
        def __init__(self):      
            self.PV = np.zeros(0)
            self.CV = np.zeros(0)
            self.SP = np.zeros(0)
            self.livetrend=0
            self.scanCount=0
        def storereads(self,CV,SP):
            self.CV=np.append(self.CV,CV)
            self.SP=np.append(self.SP,SP)
        def storepv(self,PV):
            self.PV=np.append(self.PV,PV)       
        def reset(self):
            self.PV = np.zeros(0)
            self.CV = np.zeros(0)
            self.SP = np.zeros(0)
            self.scanCount=0

    def fopdtsetup():
        process.Gain=float(modelgain.get())
        process.TimeConstant=float(modeltc.get())*10
        process.DeadTime=float(modeldt.get())*10
        process.Bias=float(ambient.get())
        process.t=0
        gData.reset()
        gData.livetrend=1
        spstatus.set("")
        pvstatus.set("")
        cvstatus.set("")
        comm.IPAddress = ip.get()
        comm.ProcessorSlot = int(slot.get()) 
        comm.SocketTimeout = 1
        button_start["state"] = "disabled"
        button_stop["state"] = "normal"
        button_trend["state"] = "normal"
        button_save["state"] = "normal"
        ip.configure(state="disabled")
        slot.configure(state="disabled")
        modelgain.configure(state="disabled")
        modeltc.configure(state="disabled")
        modeldt.configure(state="disabled")
        ambient.configure(state="disabled")
            
    def thread_start():    
        global looper
        looper = g.PeriodicInterval(start, 0.1)

    def start():
        try:
            ret = comm.Read([cvtag.get(),sptag.get()])
            if ret[0].Status=='Success':
                cvtext.set(round(ret[0].Value,2))         
                cvtag.configure(state="disabled")       
                actualcv=ret[0].Value
            else:
                cvstatus.set(ret[0].Status)
                if gData.CV.size>1:
                    actualcv=gData.CV[-1]
                else:
                    actualcv=0

            if ret[1].Status=='Success':
                sptext.set(round(ret[1].Value,2))         
                sptag.configure(state="disabled") 
                actualsp=ret[1].Value
            else:
                spstatus.set(ret[1].Status)
                if gData.SP.size>1:
                    actualsp=gData.SP[-1]
                else:
                    actualsp=0

            #Send CV to Process
            process.CV=gData.CV
            #Store Data when it is read
            gData.storereads(actualcv,actualsp)
            ts=[gData.scanCount,gData.scanCount+1]

            #Get new PV value
            if gData.PV.size>1:
                pv=process.update(gData.PV[-1],ts)
            else:
                pv=process.update(float(ambient.get()),ts)
            #Add Noise between -0.1 and 0.1
            noise=(random.randint(0,10)/100)-0.05
            #Store PV
            gData.storepv(pv[0]+noise)
            #Write PV to PLC   
            write = comm.Write(pvtag.get(),gData.PV[-1])
            if write.Status=='Success':
                pvtext.set(round(write.Value,2))         
                pvtag.configure(state="disabled")            
            else:
                pvstatus.set(write.Status)
            gData.scanCount+=1
            
        except Exception as e:
            pvstatus.set('An exception occurred: {}'.format(e))     
        
    def stop():
        button_start["state"] = "normal"
        button_stop["state"] = "disabled"
        ip.configure(state="normal")
        slot.configure(state="normal")
        modelgain.configure(state="normal")
        modeltc.configure(state="normal")
        modeldt.configure(state="normal")
        pvtag.configure(state="normal")
        cvtag.configure(state="normal")
        sptag.configure(state="normal")
        ambient.configure(state="normal")
        gData.livetrend=0
        if 'looper' in globals():
                looper.stop()
        comm.Close()
        
    def livetrend(): 
        #Set up the figure
        fig = plt.figure()
        ax = plt.axes(xlim=(0,100),ylim=(0, 100))
        SP, = ax.plot([], [], lw=2, color="Red", label='SP')
        CV, = ax.plot([], [], lw=2, color="Green", label='CV')
        PV, = ax.plot([], [], lw=2, color="Blue", label='PV')

        #Setup Func
        def init():
            SP.set_data([], [])
            PV.set_data([], [])
            CV.set_data([], [])        
            plt.ylabel('EU')  
            plt.xlabel("Time (min)")
            plt.suptitle("Live Data")        
            plt.legend(loc='best')
            return SP,PV,CV,

        #Loop here
        def animate(i):     
            x = np.arange(len(gData.SP),dtype=int)
            scale = int(60*1000/100) #Convert mS to Minutes
            x=x/scale
            ax.set_xlim(0,max(x)*1.1)
            max_y=max(max(gData.PV),max(gData.CV),max(gData.SP))
            min_y=min(min(gData.PV),min(gData.CV),min(gData.SP))
            ax.set_ylim(min_y-1,max_y+1)
            SP.set_data(x,gData.SP)
            CV.set_data(x,gData.CV)
            PV.set_data(x,gData.PV)
            return SP,PV,CV,

        #Live Data
        if gData.livetrend:
            anim = animation.FuncAnimation(fig, animate, init_func=init, frames=100, interval=1000) #, blit=True) # cant use blit with dynamic x-axis

        plt.show()
    
    def save():            
        if len(gData.PV):
            now=time.strftime("%Y%m%d_%H%M%S")
            np.savetxt("SimData_" + now + ".csv", np.transpose((gData.PV,gData.CV,gData.SP)), header="PV;CV;SP", comments='', fmt='%1.3f', delimiter = ";")
        else:
            button_save["state"] = "disabled"
    #Gui
    root = tk.Tk()
    root.title('CLX PID Process Simulator using a FOPDT Model')
    root.resizable(True, True)
    root.geometry('500x325')

    #Text tags setup
    pvtext = tk.StringVar()
    cvtext = tk.StringVar()
    sptext = tk.StringVar()
    pvstatus = tk.StringVar()
    cvstatus = tk.StringVar()
    spstatus = tk.StringVar()
    pvtag = tk.Entry(root,width=10)
    cvtag = tk.Entry(root,width=10)
    sptag = tk.Entry(root,width=10)
    ip = tk.Entry(root,width=15)
    slot = tk.Entry(root,width=5)
    modelgain = tk.Entry(root,width=5)
    modeltc = tk.Entry(root,width=5)
    modeldt = tk.Entry(root,width=5)
    ambient = tk.Entry(root,width=5)

    #Column 0
    #Labels
    tk.Label(root, text="Tag").grid(row=0,column=0,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="SP:").grid(row=1,column=0,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="PV:").grid(row=2,column=0,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="CV:").grid(row=3,column=0,padx=10 ,pady=2,sticky="NESW")
    #Row 4 = Button
    #Row 5 = Button
    tk.Label(root, text="PLC IP Address:").grid(row=6,column=0,padx=10 ,pady=2)
    tk.Label(root, text="PLC Slot:").grid(row=7,column=0,padx=10 ,pady=2)
    tk.Label(root, text="Model Gain:").grid(row=8,column=0,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="Model TimeConstant(s):").grid(row=9,column=0,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="Model DeadTime(s):").grid(row=10,column=0,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="Model Ambient:").grid(row=11,column=0,padx=10 ,pady=2,sticky="NESW")

    #Column 1
    #Labels
    tk.Label(root, text="Value").grid(row=0,column=1,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=sptext).grid(row=1,column=1,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=pvtext).grid(row=2,column=1,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=cvtext).grid(row=3,column=1,padx=10 ,pady=2,sticky="NESW")
    #Row 4 = Button
    #Row 5 = Button
    ip.grid(row=6, column=1,padx=10 ,pady=2,sticky="NESW")
    slot.grid(row=7, column=1,padx=10 ,pady=2,sticky="NESW")
    modelgain.grid(row=8, column=1,padx=10 ,pady=2,sticky="NESW")
    modeltc.grid(row=9, column=1,padx=10 ,pady=2,sticky="NESW")
    modeldt.grid(row=10, column=1,padx=10 ,pady=2,sticky="NESW")
    ambient.grid(row=11, column=1,padx=10 ,pady=2,sticky="NESW")

    #Column 2
    #Actual PLC TagName
    tk.Label(root, text="PLC Tag").grid(row=0,column=2,padx=10 ,pady=2)
    sptag.grid(row=1, column=2,padx=10 ,pady=2,sticky="NESW")
    pvtag.grid(row=2, column=2,padx=10 ,pady=2,sticky="NESW")
    cvtag.grid(row=3, column=2,padx=10 ,pady=2,sticky="NESW")

    #Column 3
    #Status
    tk.Label(root, text="Last Error:").grid(row=0,column=3,padx=10,columnspan=3 ,pady=2,sticky="W")
    tk.Label(root, textvariable=spstatus).grid(row=1,column=3,padx=10,columnspan=3 ,pady=2,sticky="W")
    tk.Label(root, textvariable=pvstatus).grid(row=2,column=3,padx=10,columnspan=3 ,pady=2,sticky="W")
    tk.Label(root, textvariable=cvstatus).grid(row=3,column=3,padx=10,columnspan=3 ,pady=2,sticky="W")

    #Default Values
    sptag.insert(10, "SP")
    pvtag.insert(10, "PID_PV")
    cvtag.insert(10, "PID_CV")
    ip.insert(10, "192.168.123.100")
    slot.insert(5, "2")
    modelgain.insert(5, "1.75")
    modeltc.insert(5, "75.5")
    modeldt.insert(5, "17.66")
    ambient.insert(5, "13.5")

    #Buttons
    #Start Button Placement
    button_start = tk.Button(root, text="Start Simulator", command=lambda :[fopdtsetup(),thread_start()])
    button_start.grid(row=4,column=0,columnspan=1,padx=10 ,pady=2,sticky="NESW")

    #Stop Button Placement
    button_stop = tk.Button(root, text="Stop Simulator", command=lambda :[stop()])
    button_stop.grid(row=4,column=1,columnspan=1,padx=10 ,pady=2,sticky="NESW")
    button_stop["state"] = "disabled"

    #Trend Button Placement
    button_trend = tk.Button(root, text="Show Trend", command=lambda :[livetrend()])
    button_trend.grid(row=5,column=0,columnspan=2,padx=10 ,pady=2,sticky="NESW")
    button_trend["state"] = "disabled"

    button_save = tk.Button(root, text="Save as CSV", command=lambda :[save()])
    button_save.grid(row=11,column=2,columnspan=1,padx=10 ,pady=2,sticky="NESW")
    button_save["state"] = "disabled"

    #default setup 
    params=0,0
    model= (modelgain.get(),modeltc.get()*10,modeldt.get()*10,13.1)
    process=g.FOPDTModel(params, model)
    gData=data()
    comm=PLC()

    root.mainloop()

if __name__ == '__main__':
    main()


from setuptools import setup, find_packages

setup(
    name='pytunelogix',
    version='1.1.17',
    license='MIT',
    author="Destination2Unknown",
    author_email='destination0b10unknown@gmail.com',
    description='PID tuner, logger and simulator',
    long_description='PID tuner, logger and simulator, with multiple tuning methods',
    packages=find_packages(),
    url='https://github.com/Destination2Unknown/pytunelogix',
    keywords='PID Tuner',
    classifiers= [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",        
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=[
          'scipy',
          'numpy',
          'matplotlib',
          'pylogix',
          'pandas',
      ],
)


from pytunelogix.pidlogger import clxlogger
from pytunelogix.stage1 import csvtuner
from pytunelogix.clxpidsim import clxsim
from pytunelogix.simulate import simulator
from kivy.app import App
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.core.window import Window
Window.size = (300, 400)

class ButtonApp(App):     
    def build(self):
        self.title = 'pytunelogix'
        button1 = Button(text="PID Logger", size_hint=(1, 1))
        button1.bind(on_press=self.callPIDLogger)
        button2 = Button(text="PID Tuner", size_hint=(1, 1))
        button2.bind(on_press=self.callCSV)
        button3 = Button(text="PID Simulator", size_hint=(1, 1))
        button3.bind(on_press=self.callsim)
        button4 = Button(text="PID CLX Sim", size_hint=(1, 1))
        button4.bind(on_press=self.callclxsim)
        boxlayout = BoxLayout(orientation='vertical')
        boxlayout.add_widget(button1)
        boxlayout.add_widget(button2)
        boxlayout.add_widget(button3)
        boxlayout.add_widget(button4)
        return boxlayout

    def callPIDLogger(self,obj):
        clxlogger.main()

    def callCSV(self,obj):
        csvtuner.main()

    def callsim(self,obj):
        simulator.main()
    
    def callclxsim(self,obj):
        clxsim.main()

if '__main__' == __name__:
    root = ButtonApp()
    root.run()


"""
   
Updated and maintained by destination0b10unknown@gmail.com
Copyright 2022 destination2unknown

Licensed under the MIT License;
you may not use this file except in compliance with the License.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
   
"""
import sys
import time
import tkinter as tk
import csv
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import threading
from matplotlib import animation
from datetime import datetime
from pylogix import PLC
from pytunelogix.common import generalclasses as g

def main():
    class data(object):
        def __init__(self):      
            self.reset()

        def update(self,PV,CV,SP):
            self.PV=np.append(self.PV,PV)
            self.CV=np.append(self.CV,CV)
            self.SP=np.append(self.SP,SP) 
            
        def reset(self):
            self.PV = np.zeros(0)
            self.CV = np.zeros(0)
            self.SP = np.zeros(0)        
            self.ErrCount=0
            self.ReadCount=0
            self.SetupFlag=False
            self.RunNowFlag=False
            self.CSVFile=object
            self.CSVFileWriter=object
                
    def thread_record():
        global loop_record
        loop_record = g.PeriodicInterval(Record, int(deltat.get())/1000)
        
    def Record():
        if GData.SetupFlag==False:
            #Setup communnication object       
            comm.IPAddress = ip.get()
            comm.ProcessorSlot = int(slot.get())
            comm.SocketTimeout = sorted([100, int(deltat.get())/1000, 5000])[1]
            spstatus.set("")
            pvstatus.set("")
            cvstatus.set("")        
            GData.ErrCount=0
            GData.ReadCount=0
            GData.SetupFlag=True 
            GData.RunNowFlag=True
            button_record.configure(bg = "Black")
            button_record["state"] = "disabled"
            try:
                #Write new data to csv if read was successful, if not write last value, Open File or create if it doesn't exist
                GData.CSVFile = open(fname.get(), 'a')
                GData.CSVFileWriter = csv.writer(GData.CSVFile, delimiter=';', lineterminator='\n', quotechar='/', quoting=csv.QUOTE_MINIMAL)
                #Write headers if its a new file
                if os.stat(fname.get()).st_size == 0:
                    GData.CSVFileWriter.writerow(('PV','CV','SP','TimeStamp'))

            except Exception as e:    
                spstatus.set('File Error: ' + str(e))  
                cvstatus.set('File Error: ' + str(e))  
                pvstatus.set('File Error: ' + str(e)) 

        current_date_time = datetime.utcnow().strftime('%d-%m-%Y %H:%M:%S.%f')    

        try:
            #Setup tags to read
            tag_list = [pvtexttag.get(), cvtexttag.get(), sptexttag.get()]
            ret = comm.Read(tag_list)
    
            #Update gui data if read is successful, if not update last error
            if ret[0].Status=='Success':
                actualpv=round(ret[0].Value,2)
                pvtext.set(actualpv)         
                pvtexttag.configure(state="disabled")
            else:
                pvstatus.set(ret[0].Status)
                if len(GData.PV>0):
                    actualpv=GData.PV[-1]
                else:
                    actualpv=0

            #Update gui data if read is successful, if not update last error
            if ret[1].Status=='Success':
                actualcv=round(ret[1].Value,2)
                cvtext.set(actualcv)         
                cvtexttag.configure(state="disabled")
            else:
                cvstatus.set(ret[1].Status)
                if len(GData.CV>0):
                    actualcv=GData.CV[-1]
                else:
                    actualcv=0

            #Update gui data if read is successful, if not update last error
            if ret[2].Status=='Success':
                actualsp=round(ret[2].Value,2)
                sptext.set(actualsp)
                sptexttag.configure(state="disabled")
            else:
                spstatus.set(ret[2].Status)
                if len(GData.SP>0):
                    actualsp=GData.SP[-1]
                else:
                    actualsp=0
                    
            #Disable inputs if tag read is successful
            if ret[0].Status=='Success' or ret[1].Status=='Success' or ret[2].Status=='Success':
                deltat.configure(state="disabled")
                ip.configure(state="disabled")
                slot.configure(state="disabled")
                fname.configure(state="disabled")
                button_record.configure(bg = "Green")
                button_livetrend["state"] = "normal"            

            #Write all values to csv file  
            row = [actualpv,actualcv,actualsp]
            GData.update(actualpv,actualcv,actualsp)            
            row.append(current_date_time)
            GData.CSVFileWriter.writerow(row)

            #If read fails update error counter, if successful increment read counter
            if ret[0].Status=='Success' or ret[1].Status=='Success' or ret[2].Status=='Success':
                GData.ReadCount+=1
            if ret[0].Status!='Success' and ret[1].Status!='Success' and ret[2].Status!='Success':
                GData.ErrCount+=1
            ec='Errors: '+ str(GData.ErrCount)
            rc='Reads: '+ str(GData.ReadCount)
            errorcount.set(ec)
            readcount.set(rc)

        except Exception as e:    
            spstatus.set('Error: ' + str(e))  
            cvstatus.set('Error: ' + str(e))  
            pvstatus.set('Error: ' + str(e))     
            
    def Write():
        try:
            #Setup comms
            comms = PLC()   
            comms.IPAddress = ip.get()
            comms.ProcessorSlot = int(slot.get())
            comms.SocketTimeout = 1

            #Setup tags to read back data
            tag_list = [cvtexttag.get(), sptexttag.get()]    
        
            #Don't write data if empty, reads back data after write
            if spsend.get(): 
                sp = float(spsend.get())    
                comms.Write(sptexttag.get(), sp)              
                sptext.set(round(comms.Read(sptexttag.get()).Value,2)) 
            if cvsend.get():
                cv=float(cvsend.get())
                comms.Write(cvtexttag.get(), cv)
                cvtext.set(round(comms.Read(cvtexttag.get()).Value,2))

        except Exception as e:    
            spstatus.set('Write Error: ' + str(e))  
            cvstatus.set('Write Error: ' + str(e))  
            
        finally:
            comms.Close()

    def TrendFileData():
        try:
            if GData.RunNowFlag:
                GData.CSVFile.flush()        
            df = pd.read_csv(fname.get(), sep=';',quoting=csv.QUOTE_NONE, escapechar="\\", encoding="utf-8")
            headers=list(df)
            df['TimeStamp'] = pd.to_datetime(df['TimeStamp'],format='%d-%m-%Y %H:%M:%S.%f')
            plt.figure()
            plt.plot(df['TimeStamp'],df[headers[0]], color="#1f77b4", linewidth=2, label=headers[0])
            plt.plot(df['TimeStamp'],df[headers[1]], color="#ff7f0e",linewidth=2,label=headers[1])
            plt.plot(df['TimeStamp'],df[headers[2]], color="#2ca02c",linewidth=2,label=headers[2])
            plt.ylabel('EU')                   
            plt.xlabel("Time")
            plt.title(fname.get())
            plt.legend(loc='best')
            plt.gcf().autofmt_xdate()
            plt.show()
        
        except Exception as e:    
            pvstatus.set('CSV Read Error: ' + str(e))  

    
    def LiveTrend(): 
        #Set up the figure
        fig = plt.figure()
        ax = plt.axes(xlim=(0,100),ylim=(0, 100))
        SP, = ax.plot([], [], lw=2, label='SP')
        PV, = ax.plot([], [], lw=2, label='PV')
        CV, = ax.plot([], [], lw=2, label='CV')

        #Setup Func
        def init():
            SP.set_data([], [])
            PV.set_data([], [])
            CV.set_data([], [])
            plt.ylabel('EU')  
            plt.xlabel("Time (min)")
            plt.suptitle("Live Data")        
            plt.legend(loc='best')
            return SP,PV,CV,

        #Loop here
        def animate(i):     
            x = np.arange(len(GData.SP),dtype=int)
            scale = int(60*1000/int(deltat.get())) #Convert mS to Minutes
            x=x/scale
            ax.set_xlim(0,max(x)*1.1)
            SP.set_data(x,GData.SP)
            CV.set_data(x,GData.CV)
            PV.set_data(x,GData.PV)
            return SP,PV,CV,

        #Live Data
        if GData.RunNowFlag:
            anim = animation.FuncAnimation(fig, animate, init_func=init, frames=100, interval=1000) #, blit=True) # cant use blit with dynamic x-axis

        plt.show()

    def Stop():        
        if 'loop_record' in globals():
            loop_record.stop()          
        #Enable text box entry
        sptexttag.configure(state="normal")
        pvtexttag.configure(state="normal")
        cvtexttag.configure(state="normal")
        deltat.configure(state="normal")
        ip.configure(state="normal")
        slot.configure(state="normal")
        fname.configure(state="normal")        
        button_record.configure(bg = "#f0f0f0")
        button_livetrend["state"] = "disabled"
        button_record["state"] = "normal"
        comm.Close()  
        if not GData.CSVFile.closed:
            GData.CSVFile.close()
        GData.reset()
        plt.close('all')

    #Gui
    root = tk.Tk()
    root.title('CLX PID Data Logger -> CSV')
    root.resizable(True, True)
    root.geometry('510x175')

    #Text tags setup
    pvtext = tk.StringVar()
    cvtext = tk.StringVar()
    sptext = tk.StringVar()
    pvstatus = tk.StringVar()
    cvstatus = tk.StringVar()
    spstatus = tk.StringVar()
    errorcount = tk.StringVar()
    readcount = tk.StringVar()
    sptexttag = tk.Entry(root,width=10)
    pvtexttag = tk.Entry(root,width=10)
    cvtexttag = tk.Entry(root,width=10)
    spsend = tk.Entry(root,width=5)
    cvsend = tk.Entry(root,width=5)
    deltat = tk.Entry(root,width=5)
    ip = tk.Entry(root,width=15)
    slot = tk.Entry(root,width=5)
    fname = tk.Entry(root,width=5)

    #Column 0
    #Labels
    tk.Label(root, text="  ").grid(row=0,column=0,padx=10 ,pady=2)
    tk.Label(root, text="Tag").grid(row=0,column=0,padx=10 ,pady=2)
    tk.Label(root, text="SP:").grid(row=1,column=0,padx=10 ,pady=2)
    tk.Label(root, text="PV:").grid(row=2,column=0,padx=10 ,pady=2)
    tk.Label(root, text="CV:").grid(row=3,column=0,padx=10 ,pady=2)

    #Column 1
    #Label positions - Read
    tk.Label(root, text="Value").grid(row=0,column=1,padx=10 ,pady=2)
    tk.Label(root, textvariable=sptext).grid(row=1,column=1,padx=10 ,pady=2)
    tk.Label(root, textvariable=pvtext).grid(row=2,column=1,padx=10 ,pady=2)
    tk.Label(root, textvariable=cvtext).grid(row=3,column=1,padx=10 ,pady=2)

    #Column 2
    #Send - Write
    tk.Label(root, text="Write").grid(row=0,column=2,padx=10 ,pady=2)
    spsend.grid(row=1, column=2,padx=10 ,pady=2,sticky="NESW")
    cvsend.grid(row=3, column=2,padx=10 ,pady=2,sticky="NESW")

    #Column 3
    #Actual PLC TagName
    tk.Label(root, text="PLC Tag").grid(row=0,column=3,padx=10 ,pady=2)
    sptexttag.grid(row=1, column=3,padx=10 ,pady=2,sticky="NESW")
    pvtexttag.grid(row=2, column=3,padx=10 ,pady=2,sticky="NESW")
    cvtexttag.grid(row=3, column=3,padx=10 ,pady=2,sticky="NESW")
    deltat.grid(row=4, column=3,columnspan=1,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="mS",bg='#F0F0F0').grid(row=4,column=3,padx=10 ,pady=2,sticky="E")

    #Column 4
    #Actual PLC IP address
    tk.Label(root, text="PLC IP Address:").grid(row=0,column=4,padx=10 ,pady=2)
    ip.grid(row=1, column=4,padx=10 ,pady=2,sticky="NESW")
    tk.Label(root, text="PLC Slot:").grid(row=2,column=4,padx=10 ,pady=2)
    slot.grid(row=3, column=4,padx=10 ,pady=2,sticky="NESW")
    fname.grid(row=4, column=4,padx=10 ,pady=2,sticky="NESW",columnspan=1)

    #Column 5
    #Status
    tk.Label(root, text="Last Error:").grid(row=0,column=5,padx=10,columnspan=2 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=spstatus).grid(row=1,column=5,padx=10,columnspan=2 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=pvstatus).grid(row=2,column=5,padx=10,columnspan=2 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=cvstatus).grid(row=3,column=5,padx=10,columnspan=2 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=errorcount).grid(row=4,column=5,padx=10,columnspan=2 ,pady=2,sticky="NESW")
    tk.Label(root, textvariable=readcount).grid(row=5,column=5,padx=10,columnspan=2 ,pady=2,sticky="NESW")

    #Default Values
    sptexttag.insert(10, "SP")
    pvtexttag.insert(10, "PID_PV")
    cvtexttag.insert(10, "PID_CV")
    deltat.insert(5, "100")
    ip.insert(10, "192.168.123.100")
    slot.insert(5, "2")
    fname.insert(10,"D:\Trend.csv")

    #Buttons
    #Record Button Placement
    button_record = tk.Button(root, text="Record Data",disabledforeground="white", command=lambda :[thread_record()])
    button_record.grid(row=4,column=0,columnspan=2,padx=10 ,pady=2,sticky="NESW")

    #Write Button Placement
    button_write = tk.Button(root, text="Write", command=lambda :[Write()])
    button_write.grid(row=4,column=2,columnspan=1,padx=10 ,pady=2,sticky="NESW")

    #Live Trend Button Placement
    button_livetrend = tk.Button(root, text="Live Plot", command=lambda :[LiveTrend()])
    button_livetrend.grid(row=5,column=3,columnspan=1,padx=10 ,pady=2,sticky="NESW")
    button_livetrend["state"] = "disabled"

    #Stop Trends Button Placement
    button_stop = tk.Button(root, text="Stop Recording", command=lambda :[Stop()])
    button_stop.grid(row=5,column=0,columnspan=3,padx=10 ,pady=2,sticky="NESW")

    #Trend Button Placement
    button_TrendFileData = tk.Button(root, text="Plot Data From CSV", command=lambda :[TrendFileData()])
    button_TrendFileData.grid(row=5,column=4,columnspan=1,padx=10 ,pady=2,sticky="NESW")

    #Class init
    GData=data()
    comm = PLC()

    root.mainloop()

if __name__ == '__main__':
    main()